/*
 * Titel:    Hauptdatei
 * Autor:    Thomas Huxhorn  web:http://www.amee-bot.de  mail:thomas.huxhorn@web.de
 * Project:  Servosteuerung
 * Hardware: ATTiny2313 8MHz
 * Version:  0.3
 * Datum: 11.03.2011
 * erste Veroeffentlichung 27.10.2005
 *
 * Servosteuerung fuer max. 8 Servos an PB0 bis PB7.
 * Ext. Oszi. an T1
 * Taster Left an PD2
 *   "    Right an PD3
 *   "    Select Servo an PD4

 * Eine 7-Segment Anzeige wird ueber den BCD Code angesteuert.
 * A = PA0
 * B = PA1
 * C = PD6
 * D = Masse
 *
 * Einstellung der Baud Rate
 * =========================
 *
 * 1. Reset und Select Servo Taste gleichzeitig druecken
 * 2. Reset Taste und erst danach Select Servo Taste loslassen
 * 3. Die Anzeige blinkt 4 mal
 * 4. Jeder Zahl ist einer Geschwindigkeit zugeordnet. Mit der Taste Rechts oder Links waehlen:
 *
 * 0 = 2400 Baud
 * 1 = 4800
 * 2 = 9600
 * 3 = 14400
 * 4 = 19200
 * 5 = 38400
 * 6 = 250000
 * 7 = 500000
 *
 * Mit einem weiteren Druck auf Select Servo wird die Einstellung uebernommen und das
 * Controllprogramm beginnt.
 *
 * Manuelle Steuerung
 * ==================
 * Mit der Select Servo Taste einen Servo auswaehlen und mit der Rechts oder Links Taster
 * seine Position veraendern
 *
 * Fernsteuerung ueber USART
 * =========================
 *
 * Die Software erwartet einen 4 Byte String.
 * 1. Byte Synbyte. Hier wird ein ";" erwartet
 * 2. Byte wie das 1. Byte
 * 3. Byte Servonummer. Hier wird das Zeichen 1 bis 8 erwartet
 * 4. Byte Hier wird die neue 8bit Servoposition erwartet
 *
 */

#include "servo5.h"

void usartChangeBaud()
{

  uint8_t i, BaudRateNr;

  BaudRateNr = 4;

  // 4 mal die Anzeige blinken lassen um zu signalisieren dass man im Setup Mode ist
  for (i = 0; i < 4; i++)
  {

    // mit ein paa Herz blinken lassen
    _delay_ms(150);
    showServoNr(0);

    _delay_ms(150);
    showServoNr(4);

  }

  while (! debounce(&PIND, SELECTSERVO))
  {


    if (debounce(&PIND, LEFT))
      BaudRateNr--;

    if (BaudRateNr == 255)
      BaudRateNr = 0;

    if (debounce(&PIND, RIGHT))
      BaudRateNr++;

    if (BaudRateNr > 7)
      BaudRateNr = 7;

    showServoNr(BaudRateNr);

  }


  switch (BaudRateNr)
  {
    case 0: UsartBaudrate =  2400;
            break;

    case 1: UsartBaudrate =  4800;
            break;

    case 2: UsartBaudrate =  9600;
            break;

    case 3: UsartBaudrate =  14400;
            break;

    case 4: UsartBaudrate =  19200;
            break;

    case 5: UsartBaudrate =  38400;
            break;

    case 6: UsartBaudrate =  250000;
            break;

    case 7: UsartBaudrate =  500000;
            break;

  } // switch

  // 3 mal die Anzeige blinken lassen um zu signalisieren dass man den Setup Mode verlassen hat
  for (i = 0; i < 2; i++)
  {

    _delay_ms(150);
    showServoNr(BaudRateNr);

    // mit 10Hz blinken lassen
    _delay_ms(150);
    showServoNr(0);

  }
}

void usartInit()
{

  UCSRB = (1<<RXEN) | (1<<TXEN) | (1 << RXCIE);

  #define USART_BAUD_SELECT (F_CPU/(UsartBaudrate *16l)-1)

  UBRRH = ((unsigned char)(USART_BAUD_SELECT>>8)) & 0x80;
  UBRRL = (unsigned char) USART_BAUD_SELECT;

}

void USART_transmit (unsigned char c)
{

  while (!(UCSRA & (1<<UDRE))) {}
  UDR = c;

}

void USART_transmit_string (unsigned char *string)
{

  while (!(UCSRA & (1<<UDRE))) {}
    while ( *string)
      USART_transmit (*string++);

}

// Neues Zeichen empfangen.Funktionen zum ueberpruefen des Zeichens aufrufen
SIGNAL(SIG_USART0_RX)
{

  unsigned char temp;

  temp = UDR;

  switch (searchForByte)
  {

    case 1: checkByte1(temp);
            break;

    case 2: checkByte2(temp);
            break;

    case 3: checkByte3(temp);
            break;

    case 4: checkByte4(temp);
            break;

  } // switch
}

// Auf das erste Synbyte testen
void checkByte1(unsigned char temp)
{

  if (temp == ';')
  {

   searchForByte = 2;
  //USART_transmit_string("OK1\n");

  }
}

// Auf das zweite Synbyte testen
void checkByte2(unsigned char temp)
{

  if (temp == ';')
  {

   searchForByte = 3;
   //USART_transmit_string("OK2\n");

  }
}

// Fuer welchen Servo ist das naechst folgende Byte bestimmt?
void checkByte3(unsigned char temp)
{

  if (temp >= '0' && temp <= '7')
  {

    stringServoNr = temp - 48;
    searchForByte  = 4;
    //USART_transmit_string("OK3\n");

  }
}

// Die gewuenschte Servoposition uebernehmen
void checkByte4(unsigned char temp)
{

  sollpos[stringServoNr] = (uint8_t) temp;
  searchForByte  = 1;
  //USART_transmit(temp);
  //USART_transmit_string(" OK4\n");

}


// Einfache Funktion zum Entprellen eines Tasters
uint8_t debounce(volatile uint8_t *port, uint8_t pin)
{

  if ( ! (*port & (1 << pin)) )
  {

    // Pin wurde auf Masse gezogen, etwa 8ms warten
_delay_ms(50);
    if ( ! (*port & (1 << pin)) )
    {

      // Anwender Zeit zum Loslassen des Tasters geben
_delay_ms(50);
      return 1;
    }
  }

  // Die Taste wurde nicht gedrueckt
  return 0;
}

// Einfache Funktion zum Entprellen eines Tasters
uint8_t debounce_fast(volatile uint8_t *port, uint8_t pin)
{

  if ( ! (*port & (1 << pin)) )
  {

    // Pin wurde auf Masse gezogen, etwa 8ms warten
_delay_ms(10);
    if ( ! (*port & (1 << pin)) )
    {

      // Anwender Zeit zum Loslassen des Tasters geben
_delay_ms(10);
      return 1;
    }
  }

  // Die Taste wurde nicht gedrueckt
  return 0;
}

// Die 3 Bits fuer die 7-Segment Anzeige sind aus Platzmangel auf
// 2 verschiedene Ports verteilt. Folgende Funktion bekommt die
// Zahl fuer die Anzeige und setze die Passenden Bits.
void showServoNr(uint8_t servoNr)
{

  if (bit_is_set(servoNr,0))
    PORTA |= (1 << SEGMENTBITA);
  else
    PORTA &= ~(1 << SEGMENTBITA);

  if (bit_is_set(servoNr,1))
    PORTA |= (1 << SEGMENTBITB);
  else
    PORTA &= ~(1 << SEGMENTBITB);

  if (bit_is_set(servoNr,2))
    PORTD |= (1 << SEGMENTBITC);
  else
    PORTD &= ~(1 << SEGMENTBITC);

}

// Den Timer fuer das naechste High am Servoport einstellen
SIGNAL(SIG_TIMER1_COMPA)
{

  uint8_t i;
  uint16_t resttime;

  switch (whichServo)
  {

    case 1: PORTB |= (1 << SERVO1);
            TCNT1 = 0;
            OCR1A = sollpos[0] + MINIMAL;
            whichServo = 2;
            break;

    case 2: PORTB &=~ (1 << SERVO1);
            PORTB |= (1 << SERVO2);
            TCNT1 = 0;
            OCR1A = sollpos[1] + MINIMAL;
            whichServo = 3;
            break;

    case 3: PORTB &=~ (1 << SERVO2);
            PORTB |= (1 << SERVO3);
            TCNT1 = 0;
            OCR1A = sollpos[2] + MINIMAL;
            whichServo = 4;
            break;

    case 4: PORTB &=~ (1 << SERVO3);
            PORTB |= (1 << SERVO4);
            TCNT1 = 0;
            OCR1A = sollpos[3] + MINIMAL;
            whichServo = 5;
            break;

    case 5: PORTB &=~ (1 << SERVO4);
            PORTB |= (1 << SERVO5);
            TCNT1 = 0;
            OCR1A = sollpos[4] + MINIMAL;
            whichServo = 6;
            break;

    case 6: PORTB &=~ (1 << SERVO5);
            PORTB |= (1 << SERVO6);
            TCNT1 = 0;
            OCR1A = sollpos[5] + MINIMAL;
            whichServo = 7;
            break;

    case 7: PORTB &=~ (1 << SERVO6);
            PORTB |= (1 << SERVO7);
            TCNT1 = 0;
            OCR1A = sollpos[6] + MINIMAL;
            whichServo = 8;
            break;

    case 8: PORTB &=~ (1 << SERVO7);
            PORTB |= (1 << SERVO8);
            TCNT1 = 0;
            OCR1A = sollpos[7] + MINIMAL;
            whichServo = 9;
            break;

    // Restliche Zeit Warten
    case 9: PORTB &=~ (1 << SERVO8);
            TCNT1 = 0;
            // Zeit zusammenzaehlen die die Servos schon verbraucht haben und
            // von 20ms abziehen. Das Ergebis lang warten, dann von vorne
            // beginnen. 146122Hz ist die Frequenz der Oszilators. 20ms geteilt
            // durch die Zeit einer Periode ergibt die Zahl die man fuer 20ms
            // warten muss. 0.02s / (1 / 146122Hz) = 2922
            // Auf diese Weise werden die 20ms nicht fuer jeden einzelnen
            // Servo genau eingehalten, aber angeblich darf die Wartezeit
            // zwischen 10 und 30ms schwanken.
            resttime = 0;
            for (i = 0; i <= 7; i++)
            {

              resttime += MINIMAL;
              resttime += sollpos[i];
            }

            OCR1A = 2922 - resttime;
            whichServo = 1;
            break;

  }// switch
}

int main()
{

  // Clk. auf 8MHz setzen
  CLKPR = 128;
  // Prescaler = 1
  CLKPR = 0;

  // PA0 und PA1 als Ausgang
  DDRA  |= (1 << PA0) | (1 << PA1);
  PORTA &= ~(1 << PA0) | (1 << PA1);

  // PORTB als Ausgang
  DDRB  = 0xFF;
  PORTB = 0x00;

  //  Nur PD6 als Ausgang
  DDRD   |= (1 << SEGMENTBITC);
  // Pullups aktivieren
  PORTD  |= (1 << LEFT) | (1 << RIGHT) | (1 << SELECTSERVO);

  // Timer1 ext. clock fallende Flanke
  TCCR1B |= (1 << CS12) | (1 << CS11);
  // Output Compare Match A Interruot aktivieren
  TIMSK  |= (1 << OCIE1A);

  TCNT1 = 0;
  OCR1A = 0;

  uint8_t selectServo = 0;

  // Welcher Servoport soll als naechstes auf HIGH gehen
  whichServo    = 1;
  // Zu welchem Servo gehoert die Position die im String geliefert wird
  stringServoNr = 0;
  // Welches Byte wir vom String untersuchen muessen
  searchForByte = 1;

  // Position die der Servo beim Einschalten haben soll
  sollpos[0] = GERADEAUS;
  sollpos[1] = GERADEAUS;
  sollpos[2] = GERADEAUS;
  sollpos[3] = GERADEAUS;
  sollpos[4] = GERADEAUS;
  sollpos[5] = GERADEAUS;
  sollpos[6] = GERADEAUS;
  sollpos[7] = GERADEAUS;

  // Standard Baudrate
  UsartBaudrate = 19200;

  if (bit_is_clear(PIND,SELECTSERVO))
    usartChangeBaud();

  usartInit();

  sei();

  for(;;) //ever
  {

    // Taste Select Servo gedrueckt?
    if (debounce(&PIND, SELECTSERVO))
    {

      selectServo++;
      if (selectServo > 7)
        selectServo = 0;

      showServoNr(selectServo);

    }

    // Taste Links gedrueckt?
    if (debounce_fast(&PIND, LEFT))
      sollpos[selectServo]+=2;

    // Taste Rechts gedrueckt?
    if (debounce_fast(&PIND, RIGHT))
      sollpos[selectServo]-=2;

  } // for ever
}
